"""
TODO: The plugin already gets imported and used, even if it isn't in __all__ here!
Is it ready to be used though? Is there a way we could allow an "opt-in" mechanism for the
plugin to be used instead?
"""
# from .auto_schema_plugin import AutoSchemaPlugin, AutoSchemaPluginConfig

__all__ = [
    # "AutoSchemaPlugin",
    # "AutoSchemaPluginConfig",
]
